//  ComputationalAssembler.java

import swarm.Globals;
import swarm.defobj.Zone;
import swarm.objectbase.SwarmObjectImpl;

import swarm.collections.ListImpl;
import swarm.collections.ListIndex;


/**
 * The ComputationalAssembler class instances make 
 * computational processes; we have a computationa assembler instance for
 * each unit
 *
 * @author Pietro Terna 
 */
public class ComputationalAssembler extends ComputationalAssemblerBasic{

    /**
     * the constructor for ComputationalAssembler
     */

    public ComputationalAssembler (Zone aZone, OrderDistiller od, ESFrameModelSwarm mo)
    {
	// Call the constructor for the parent class.
	super(aZone, od, mo);

    }

    /** drop the computational assembler */
    public void drop(){

	InformationRepository.addToHibernatedComputationalAssemblerList(this);
    }

    /** computational operations with code -12xx (to be kept in 
     *  ComputationalAssembler and not to be transferred in 
     *  ComputationalAssemblerBasic) <br><br>
     *
     */



    /** modify memory matrixes
     *  <br><br>
     *  the computational step 1220 (a code for the workers_skills_firms 
     *  application) acts via 1120 code
     *
     *  it modifies the coordinates of the memory matrixes where to apply 
     *  the addenda, both in the unit the order is in and in the unit the 
     *  order comes from. In this case the displacement in the memory matrixes 
     *  is zero, but as a trace for future modifications, this feature is fully 
     *  developed. The four coordinates (two for the memory matrix of the unit 
     *  the order is in and two or the unit the order comes from, are in 
     *  positions 0,2; 0,3; 0,4, 0,5 of the second received matrix
     */
    public void c1220(){

	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=2)
	    {
		System.out.println("Code -1220 requires 2 matrixes; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }

	rd=4;

	// displacements for the unit memory matrixes coordinates
	mm1=(MemoryMatrix) pendingComputationalSpecificationSet.
	    getMemoryMatrixAddress(1);
	layer=pendingComputationalSpecificationSet.
	    getOrderLayer();
	urd0=(int) mm1.getValue(layer, 0+rd, 2);
	ucd0=(int) mm1.getValue(layer, 0+rd, 3);
	urd1=(int) mm1.getValue(layer, 0+rd, 4);
	ucd1=(int) mm1.getValue(layer, 0+rd, 5);

	checkMatrixNumber=false;
	c1120();
	checkMatrixNumber=true;
	rd=0;
	urd0=0; ucd0=0; urd1=0; ucd1=0;

    } // end c1220

    /** unit display on another stratum
     *  <br><br>
     *  the computational step 1250 (a code for the workers_skills_firms 
     *  application) acts via 1110 code
     *
     *  when received by one unit in a stratum, acts vs. all the units of the same 
     *  type (i.e., with the same production phase) in all strata (excluding those 
     *  of the destination stratum), copying them on the destination stratum
     *
     *  finally it changes the status to done; the change is also done in 1110, but in 
     *  case non similar units were found, ww have to repeat it there
     */
    public void c1250(){
	int i, im, thePhase, unitCount, layer=0;
	Unit theUnitTheOrderIsIn;
	Order theCurrentOrder;
	ListImpl aList;
	MemoryMatrix mU;


	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=1 && checkMatrixNumber)
	    {
		System.out.println("Code -1250 requires one matrix; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }

	theCurrentOrder=pendingComputationalSpecificationSet.getOrder();
	theUnitTheOrderIsIn=theCurrentOrder.getTheUnitTheOrderIsIn();
	thePhase=theUnitTheOrderIsIn.getProductionPhase();
	
	ucd=1; // the actual position of the control cell for c1110 in the memory 
	       // matrix of the unit is displaced of 1 column
	for (im=0;im<InformationRepository.getStratumNumber();im++)
	    {
		aList=(ListImpl) InformationRepository.getUnitListFromModelSequence(im);
		unitCount=aList.getCount();
		for (i=0; i<unitCount;i++)
		    {
			// whichUnit is substituting, in c1110, theUnitTheOrderIsIn
			// because here we apply c1110 to a list
			whichUnit=(Unit) aList.atOffset(i);
			// is the unit showing the same production phase of the
			// receiving unit?
			if(thePhase==whichUnit.getProductionPhase())
			    {
				mU=whichUnit.getMemoryMatrix();
				if(! whichUnit.getDroppable())
				    mU.setValue(layer,0,1,1.0);
				c1111();
			    }
		    }
	    }

	ucd=0;
	whichUnit=null;
	done=true; // in case no c1110 step was executed

    } // end c1250


    /** unit display on another stratum, case of units with phase 1
     *  <br><br>
     *  the computational step 1251 (a code for the workers_skills_firms 
     *  application) acts via 1250 code, setting to 1 the row used 
     *  in matrix 0
     */
    public void c1251(){
	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=1 && checkMatrixNumber)
	    {
		System.out.println("Code -1251 requires one matrix; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }
	rd=1;
	c1250();
	rd=0;
    } // end c1251

    /** unit display on another stratum, case of units with phase 2
     *  <br><br>
     *  the computational step 1252 (a code for the workers_skills_firms 
     *  application) acts via 1250 code, setting to 2 the row used 
     *  in matrix 0
     */
    public void c1252(){
	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=1 && checkMatrixNumber)
	    {
		System.out.println("Code -1252 requires one matrix; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }
	rd=2;
	c1250();
	rd=0;
    } // end c1252

    /** unit display on another stratum, case of units with phase 3
     *  <br><br>
     *  the computational step 1253 (a code for the workers_skills_firms 
     *  application) acts via 1250 code, setting to 3 the row used 
     *  in matrix 0
     */
    public void c1253(){
	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=1 && checkMatrixNumber)
	    {
		System.out.println("Code -1253 requires one matrix; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }
	rd=3;
	c1250();
	rd=0;
    } // end c1253

    /** unit display on another stratum, case of units with phase 1001
     *  <br><br>
     *  the computational step 1261 (a code for the workers_skills_firms 
     *  application) acts via 1260 code, setting to 3 the row used 
     *  in matrix 0
     */
    public void c1261(){
	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=1 && checkMatrixNumber)
	    {
		System.out.println("Code -1261 requires one matrix; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }
	rd=0;
	c1260();
	rd=0;
    } // end c1261

   /** unit display on another stratum, case of units with phase 1002
     *  <br><br>
     *  the computational step 1262 (a code for the workers_skills_firms 
     *  application) acts via 1260 code, setting to 3 the row used 
     *  in matrix 0
     */
    public void c1262(){
	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=1 && checkMatrixNumber)
	    {
		System.out.println("Code -1262 requires one matrix; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }
	rd=1;
	c1260();
	rd=0;
    } // end c1262

   /** unit display on another stratum, case of units with phase 1003
     *  <br><br>
     *  the computational step 1263 (a code for the workers_skills_firms 
     *  application) acts via 1260 code, setting to 3 the row used 
     *  in matrix 0
     */
    public void c1263(){
	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=1 && checkMatrixNumber)
	    {
		System.out.println("Code -1263 requires one matrix; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }
	rd=2;
	c1260();
	rd=0;
    } // end c1263

    /** unit display on another stratum
     *  <br><br>
     *  the computational step 1260 (a code for the workers_skills_firms 
     *  application) acts via 1110 code
     *
     *  when received by one unit in a stratum, acts vs. all the units of the same 
     *  type (i.e., with the same production phase) in all strata (excluding those 
     *  of the destination stratum), copying them on the destination stratum
     *
     *  finally it changes the status to done; the change is also done in 1110, but in 
     *  case non similar units were found, ww have to repeat it there
     */
    public void c1260(){
	int i, im, thePhase, unitCount, layer=0;
	double threshold;
	Unit theUnitTheOrderIsIn;
	Order theCurrentOrder;
	ListImpl aList;
	MemoryMatrix mU;


	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=1 && checkMatrixNumber)
	    {
		System.out.println("Code -1260 requires one matrix; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }

	mm0=(MemoryMatrix) pendingComputationalSpecificationSet.
	    getMemoryMatrixAddress(0);
	layer=pendingComputationalSpecificationSet.
	    getOrderLayer();

	threshold = mm0.getValue(layer,0+rd,3+cd);

	theCurrentOrder=pendingComputationalSpecificationSet.getOrder();
	theUnitTheOrderIsIn=theCurrentOrder.getTheUnitTheOrderIsIn();
	thePhase=theUnitTheOrderIsIn.getProductionPhase();
	
	ucd=1; // the actual position of the control cell for c1110 in the memory 
	       // matrix of the unit is displaced of 1 column
	for (im=0;im<InformationRepository.getStratumNumber();im++)
	    {
		aList=(ListImpl) InformationRepository.getUnitListFromModelSequence(im);
		unitCount=aList.getCount();
		for (i=0; i<unitCount;i++)
		    {
			// whichUnit is substituting, in c1110, theUnitTheOrderIsIn
			// because here we apply c1110 to a list
			whichUnit=(Unit) aList.atOffset(i);
			// is the unit showing the same production phase of the
			// receiving unit?
			if(thePhase==whichUnit.getProductionPhase())
			    {
				mU=whichUnit.getMemoryMatrix();
				if(! whichUnit.getDroppable()

				   && // visibility threshold
				   
				   whichUnit.getVisibility()>threshold
				   )
				    mU.setValue(layer,0,1,1.0);
				c1111();
			    }
		    }
	    }

	ucd=0;
	whichUnit=null;
	done=true; // in case no c1110 step was executed

    } // end c1260

    /** this code simply call the highly usable c1199 computational code, using rd=5 to direct
     *  the called step to the correct row of the input matrixes 
     */
    public void c1297(){

	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=2)
	    {
		System.out.println("Code -1297 requires 2 matrixes; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }

	rd=5;
	c1199();
	rd=0;

    } // end c1297

    /** this code simply call the highly usable c1199 computational code, using rd=6 to direct
     *  the called step to the correct row of the input matrixes 
     */
    public void c1298(){

	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=2)
	    {
		System.out.println("Code -1298 requires 2 matrixes; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }

	rd=6;
	c1199();
	rd=0;

    } // end c1298

    /** this code simply call the highly usable c1199 computational code, using rd=7 to direct
     *  the called step to the correct row of the input matrixes 
     */
    public void c1299(){

	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=2)
	    {
		System.out.println("Code -1299 requires 2 matrixes; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }

	rd=7;
	c1199();
	rd=0;

    } // end c1299


}
