//AssigningTool.java

import swarm.Globals;
import swarm.defobj.Zone;
import swarm.objectbase.SwarmObjectImpl;

import swarm.collections.ListImpl;
import swarm.collections.ListIndex;
import swarm.collections.ArrayImpl;
import swarm.collections.ListShufflerImpl;

/**
 * A tool to choose the unit to which we have to send an order
 *
 * @author Pietro Terna
 */
public class AssigningTool extends SwarmObjectImpl
{
    /** the list of the normal units existing in our environment*/
    public ListImpl unitList;
    /** the iterator of unitList */
    public ListIndex unitListIndex;
    /** the temporary address of an existing unit */
    public Unit aUnit;
    /** the activity space of the first unit */
    public UnitActivitySpace unitActivitySpace;
    /** the min level of inter-visibility; below this level two units cannot
     *  exchange products; this value is not used in assignment coming from 
     *  OrderGenerator; this value can be different from one model to another: that
     *  of the sending unit is used */
    public double interVisibilityMinLevel;
    /** copy of the unit list, to be locally shuffled */
    public ListImpl myUnitList;
    /** the iterator of myUnitList */
    public ListIndex myUnitListIndex;
    /** a shuffler for the list */
    public ListShufflerImpl myShuffler;
    /** the global used unit list, employed if uniqueAssignmentInEachCycle is true  */
    public ListImpl myUsedUnitList;
    /** the reference time used by the uniqueAssignmentInEachCycle strategy */
    public float referenceTime=-1;

    /** the numner of our stratum and the number of strata (or different models) 
     *  in this run */
    public int stratum, stratumNumber;

    /** the constructor */
    public AssigningTool (Zone aZone, int straN, int stra
			  /*ListImpl ul, ListIndex uli, 
			    UnitActivitySpace uas, double iv*/
			  )
    {
	super(aZone);

	// the number of strata (differen models) in this run
	stratumNumber=straN;
	// the stratum of our model
	stratum=stra;
	// local copy of the lists
	myUnitList = new ListImpl(getZone());
	// the global used unit list
	myUsedUnitList = new ListImpl(getZone());
	// and its iterator
	myUnitListIndex = myUnitList.listBegin(getZone());
	// the shuffler
	myShuffler= new ListShufflerImpl(getZone());

    }

    /** assignment between two units */
    public Order assign(Unit sendingUnit, Order firstOrder)
    {
	int i, ii, stra, stratumSendingUnit, stratumReceivingUnit;
	boolean unitNotFound, applied;

	// new cycle?
	if(referenceTime != Globals.env.getCurrentTime())
	    {
		referenceTime=Globals.env.getCurrentTime();
		myUsedUnitList.removeAll();
	    }
	// data of the sending unit
	stratumSendingUnit=((ESFrameModelSwarm) sendingUnit.getUnitModel()).getStratum();
	unitActivitySpace = (UnitActivitySpace) InformationRepository.
	    getUnitActivitySpaceFromModel(stratumSendingUnit);
	interVisibilityMinLevel =  InformationRepository.getInterVisibilityMinLevel(stratumSendingUnit);

	// a shortcut in case of a zero time step that uses the same production phase or the unit 
	// the order is in; the order is LIFO reassigned to the same unit and executed, 
	// regardless the uniqueAssignmentInEachCycle value
	if(firstOrder.checkZeroTimeProduction() && 
	   firstOrder.getNextStep() ==	sendingUnit.getProductionPhase())
	    {
		firstOrder.setTheUnitTheOrderIsIn(sendingUnit); // this instruction is not relevant to 
		                                                // change the current status of the 
		                                                // order, but the previous one

		sendingUnit.setWaitingList(firstOrder); // being a zero time order it will be assigned
                                                        // LIFO via setWaitingList and immediately
		                                        // executed in the unit step1 cycle

		myUsedUnitList.addLast(sendingUnit);    // in case the unit has not just been kept in this 
		                                        // list in the current cycle
		return (Order) null;
	    }

	unitNotFound=true;
	for (stra=0;stra<stratumNumber && unitNotFound;stra++)
	    {
		unitList= (ListImpl) InformationRepository.
		    getUnitListFromReorderedModelSequence(stra);
		unitListIndex= (ListIndex) InformationRepository.
		    getUnitListIndexFromReorderedModelSequence(stra);


		// voiding myUnitList
		if(myUnitList.getCount()>0)myUnitList.removeAll();

		// coying and (if necessary) shuffling the list
		if(unitList.getCount()>0)unitListIndex.setOffset(0);
		for (i=0;i<unitList.getCount();i++)
		    {
			myUnitList.addLast((Unit)unitListIndex.get());
			unitListIndex.next();
		    }
		if(InformationRepository.getShuffleListsAtEachAssignment(stratumSendingUnit))
		    myShuffler.shuffleWholeList(myUnitList);

		// using the local list (shuffled, if any)
		if(myUnitList.getCount()>0)myUnitListIndex.setOffset(0);
		for (i=0;i<myUnitList.getCount() && unitNotFound;i++)
		    {

			aUnit=(Unit)myUnitListIndex.get();
			stratumReceivingUnit=((ESFrameModelSwarm) aUnit.
					      getUnitModel()).getStratum();

			if((! InformationRepository.
			    getUniqueAssignmentInEachCycle(stratumReceivingUnit))
			   || (InformationRepository.
			       getUniqueAssignmentInEachCycle(stratumReceivingUnit) &&
			       (! myUsedUnitList.contains(aUnit))))
			    {
				if (firstOrder.getNextStep() ==	aUnit.getProductionPhase())
				    {
					// no visibility check or same unit 
					if (interVisibilityMinLevel == 0 || sendingUnit == aUnit)
					    unitNotFound = false;
					// second unit belongs to the first unit area
					else if (aUnit.getXPos()<unitActivitySpace.getXSize() &&
						 aUnit.getYPos()<unitActivitySpace.getYSize())
					    //units seeing each other above the min visibility level
					    if(unitActivitySpace.interVisibility(sendingUnit, aUnit) >= 
					       interVisibilityMinLevel)unitNotFound = false;
				    }
			    }
			myUnitListIndex.next();
		    }
	    }
	if(! unitNotFound)
	    {
		firstOrder.setTheUnitTheOrderIsIn(aUnit);
		aUnit.setWaitingList(firstOrder);
		myUsedUnitList.addLast(aUnit);
		return (Order) null;
	    }
	else return firstOrder;

    }

    /** assignment from the order generator or the order distiller*/
    public void assign(Order firstOrder)
    {
	int i, ii, stra, stratumReceivingUnit;
	boolean unitNotFound, applied;

	// new cycle?
	if(referenceTime != Globals.env.getCurrentTime())
	    {
		referenceTime=Globals.env.getCurrentTime();
		myUsedUnitList.removeAll();
	    }

	unitNotFound=true;
	for (stra=0;stra<stratumNumber && unitNotFound;stra++)
	    {
		unitList= (ListImpl) InformationRepository.
		    getUnitListFromReorderedModelSequence(stra);
		unitListIndex= (ListIndex) InformationRepository.
		    getUnitListIndexFromReorderedModelSequence(stra);

		// voiding myUnitList
		if(myUnitList.getCount()>0)myUnitList.removeAll();
		// coying and (if necessary) shuffling the list
		if(unitList.getCount()>0)unitListIndex.setOffset(0);
		for (i=0;i<unitList.getCount();i++)
		    {
			myUnitList.addLast((Unit)unitListIndex.get());
			unitListIndex.next();
		    }
		if(InformationRepository.getShuffleListsAtEachAssignment(stratum))
		    // here stratum is the same of that of the OrderGenerator or of the
		    // OrderDistiller; so it has the same effect of stratumSendingUnit above
		    myShuffler.shuffleWholeList(myUnitList);

		// using the local list (shuffled, if any)
		if(myUnitList.getCount()>0)myUnitListIndex.setOffset(0);
		for (i=0;i<myUnitList.getCount() && unitNotFound ;i++)
		    {
			aUnit=(Unit)myUnitListIndex.get();
			stratumReceivingUnit=((ESFrameModelSwarm) aUnit.
					      getUnitModel()).getStratum();

			if((! InformationRepository.
			    getUniqueAssignmentInEachCycle(stratumReceivingUnit))
			   || (InformationRepository.
			       getUniqueAssignmentInEachCycle(stratumReceivingUnit) &&
			       (! myUsedUnitList.contains(aUnit))))
			    {
				if (firstOrder.getNextStep() ==	aUnit.getProductionPhase())
				    unitNotFound=false;
				
			    }
			myUnitListIndex.next();
		    }
	    }
	if(! unitNotFound)
	    {
		firstOrder.setTheUnitTheOrderIsIn(aUnit);
		aUnit.setWaitingList(firstOrder);
		myUsedUnitList.addLast(aUnit);
	    }
	else firstOrder.drop();


    }
}
